@@ -1,14 +1,18 @@
 # postgresql/ext.py
-# Copyright (C) 2005-2019 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2022 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
-# the MIT License: http://www.opensource.org/licenses/mit-license.php
+# the MIT License: https://www.opensource.org/licenses/mit-license.php
 
 from .array import ARRAY
+from ... import util
+from ...sql import coercions
 from ...sql import elements
 from ...sql import expression
 from ...sql import functions
+from ...sql import roles
+from ...sql import schema
 from ...sql.schema import ColumnCollectionConstraint
 
 
@@ -19,7 +23,7 @@ class aggregate_order_by(expression.ColumnElement):
 
         from sqlalchemy.dialects.postgresql import aggregate_order_by
         expr = func.array_agg(aggregate_order_by(table.c.a, table.c.b.desc()))
-        stmt = select([expr])
+        stmt = select(expr)
 
     would represent the expression::
 
@@ -31,7 +35,7 @@ class aggregate_order_by(expression.ColumnElement):
             table.c.a,
             aggregate_order_by(literal_column("','"), table.c.a)
         )
-        stmt = select([expr])
+        stmt = select(expr)
 
     Would represent::
 
@@ -43,23 +47,29 @@ class aggregate_order_by(expression.ColumnElement):
 
     .. seealso::
 
-        :class:`.array_agg`
+        :class:`_functions.array_agg`
 
     """
 
     __visit_name__ = "aggregate_order_by"
 
+    stringify_dialect = "postgresql"
+    inherit_cache = False
+
     def __init__(self, target, *order_by):
-        self.target = elements._literal_as_binds(target)
+        self.target = coercions.expect(roles.ExpressionElementRole, target)
+        self.type = self.target.type
 
         _lob = len(order_by)
         if _lob == 0:
             raise TypeError("at least one ORDER BY element is required")
         elif _lob == 1:
-            self.order_by = elements._literal_as_binds(order_by[0])
+            self.order_by = coercions.expect(
+                roles.ExpressionElementRole, order_by[0]
+            )
         else:
             self.order_by = elements.ClauseList(
-                *order_by, _literal_as_text=elements._literal_as_binds
+                *order_by, _literal_as_text_role=roles.ExpressionElementRole
             )
 
     def self_group(self, against=None):
@@ -80,16 +90,19 @@ class aggregate_order_by(expression.ColumnElement):
 class ExcludeConstraint(ColumnCollectionConstraint):
     """A table-level EXCLUDE constraint.
 
-    Defines an EXCLUDE constraint as described in the `postgres
+    Defines an EXCLUDE constraint as described in the `PostgreSQL
     documentation`__.
 
-    __ http://www.postgresql.org/docs/9.0/static/sql-createtable.html#SQL-CREATETABLE-EXCLUDE
+    __ https://www.postgresql.org/docs/current/static/sql-createtable.html#SQL-CREATETABLE-EXCLUDE
 
     """  # noqa
 
     __visit_name__ = "exclude_constraint"
 
     where = None
+    inherit_cache = False
+
+    create_drop_stringify_dialect = "postgresql"
 
     @elements._document_text_coercion(
         "where",
@@ -105,10 +118,12 @@ class ExcludeConstraint(ColumnCollectionConstraint):
             const = ExcludeConstraint(
                 (Column('period'), '&&'),
                 (Column('group'), '='),
-                where=(Column('group') != 'some group')
+                where=(Column('group') != 'some group'),
+                ops={'group': 'my_operator_class'}
             )
 
-        The constraint is normally embedded into the :class:`.Table` construct
+        The constraint is normally embedded into the :class:`_schema.Table`
+        construct
         directly, or added later using :meth:`.append_constraint`::
 
             some_table = Table(
@@ -123,7 +138,8 @@ class ExcludeConstraint(ColumnCollectionConstraint):
                     (some_table.c.period, '&&'),
                     (some_table.c.group, '='),
                     where=some_table.c.group != 'some group',
-                    name='some_table_excl_const'
+                    name='some_table_excl_const',
+                    ops={'group': 'my_operator_class'}
                 )
             )
 
@@ -131,11 +147,14 @@ class ExcludeConstraint(ColumnCollectionConstraint):
 
           A sequence of two tuples of the form ``(column, operator)`` where
           "column" is a SQL expression element or a raw SQL string, most
-          typically a :class:`.Column` object, and "operator" is a string
+          typically a :class:`_schema.Column` object,
+          and "operator" is a string
           containing the operator to use.   In order to specify a column name
-          when a  :class:`.Column` object is not available, while ensuring
+          when a  :class:`_schema.Column` object is not available,
+          while ensuring
           that any necessary quoting rules take effect, an ad-hoc
-          :class:`.Column` or :func:`.sql.expression.column` object should be
+          :class:`_schema.Column` or :func:`_expression.column`
+          object should be
           used.
 
         :param name:
@@ -158,6 +177,19 @@ class ExcludeConstraint(ColumnCollectionConstraint):
           If set, emit WHERE <predicate> when issuing DDL
           for this constraint.
 
+        :param ops:
+          Optional dictionary.  Used to define operator classes for the
+          elements; works the same way as that of the
+          :ref:`postgresql_ops <postgresql_operator_classes>`
+          parameter specified to the :class:`_schema.Index` construct.
+
+          .. versionadded:: 1.3.21
+
+          .. seealso::
+
+            :ref:`postgresql_operator_classes` - general description of how
+            PostgreSQL operator classes are specified.
+
         """
         columns = []
         render_exprs = []
@@ -166,7 +198,10 @@ class ExcludeConstraint(ColumnCollectionConstraint):
         expressions, operators = zip(*elements)
 
         for (expr, column, strname, add_element), operator in zip(
-            self._extract_col_expression_collection(expressions), operators
+            coercions.expect_col_expression_collection(
+                roles.DDLConstraintColumnRole, expressions
+            ),
+            operators,
         ):
             if add_element is not None:
                 columns.append(add_element)
@@ -177,8 +212,6 @@ class ExcludeConstraint(ColumnCollectionConstraint):
                 # backwards compat
                 self.operators[name] = operator
 
-            expr = expression._literal_as_column(expr)
-
             render_exprs.append((expr, name, operator))
 
         self._render_exprs = render_exprs
@@ -193,12 +226,32 @@ class ExcludeConstraint(ColumnCollectionConstraint):
         self.using = kw.get("using", "gist")
         where = kw.get("where")
         if where is not None:
-            self.where = expression._literal_as_text(
-                where, allow_coercion_to_text=True
-            )
+            self.where = coercions.expect(roles.StatementOptionRole, where)
 
-    def copy(self, **kw):
-        elements = [(col, self.operators[col]) for col in self.columns.keys()]
+        self.ops = kw.get("ops", {})
+
+    def _set_parent(self, table, **kw):
+        super(ExcludeConstraint, self)._set_parent(table)
+
+        self._render_exprs = [
+            (
+                expr if isinstance(expr, elements.ClauseElement) else colexpr,
+                name,
+                operator,
+            )
+            for (expr, name, operator), colexpr in util.zip_longest(
+                self._render_exprs, self.columns
+            )
+        ]
+
+    def _copy(self, target_table=None, **kw):
+        elements = [
+            (
+                schema._copy_expression(expr, self.parent, target_table),
+                self.operators[expr.name],
+            )
+            for expr in self.columns
+        ]
         c = self.__class__(
             *elements,
             name=self.name,
@@ -212,9 +265,9 @@ class ExcludeConstraint(ColumnCollectionConstraint):
 
 
 def array_agg(*arg, **kw):
-    """PostgreSQL-specific form of :class:`.array_agg`, ensures
-    return type is :class:`.postgresql.ARRAY` and not
-    the plain :class:`.types.ARRAY`, unless an explicit ``type_``
+    """PostgreSQL-specific form of :class:`_functions.array_agg`, ensures
+    return type is :class:`_postgresql.ARRAY` and not
+    the plain :class:`_types.ARRAY`, unless an explicit ``type_``
     is passed.
 
     .. versionadded:: 1.1
